-module(genAcs).

-export([
   main/1
   , genTree/1
]).

main(Args) ->
   [SWFile, WriteDir] = Args,
   case file:open(SWFile, [read, raw, binary, {read_ahead, 65536}, {encoding, utf8}]) of
      {ok, IoDevice} ->
         {Goto, Output} = dealEveryLine(IoDevice, _Goto = #{0 => #{}}, _Output = #{}, _State = 0),
         Failure = genFailure(Goto),
         genErl(WriteDir, Goto, Failure, Output);
      _Err ->
         io:format("genAcs open the file:~p error ~p~n", [SWFile, _Err])
   end.

dealEveryLine(IoDevice, Goto, Output, MaxState) ->
   case file:read_line(IoDevice) of
      {ok, DataStr} ->
         BinStr = binary:part(DataStr, 0, byte_size(DataStr) - 1),
         case BinStr =/= <<>> of
            true ->
               {NewGoto, NewState, NewMaxState} = addPattern(BinStr, Goto, 0, MaxState),
               NewOutput = Output#{NewState => BinStr},
               dealEveryLine(IoDevice, NewGoto, NewOutput, NewMaxState);
            _ ->
               dealEveryLine(IoDevice, Goto, Output, MaxState)
         end;
      eof ->
         {Goto, Output};
      _Err ->
         io:format("genAcs read the file error ~p~n", [_Err])
   end.

%% 从字符串模式列表构建ac搜索树
genTree(BinStrList) ->
   %% 先构造 goto and output table
   {Goto, Output} = genGotoOutput(BinStrList, _Goto = #{0 => #{}}, _Output = #{}, _State = 0),
   %% 然后构造 failure table
   Failure = genFailure(Goto),
   {Goto, Failure, Output}.

genGotoOutput([BinStr | Tail], Goto, Output, MaxState) ->
   case BinStr =/= <<>> of
      true ->
         {NewGoto, NewState, NewMaxState} = addPattern(BinStr, Goto, 0, MaxState),
         NewOutput = Output#{NewState => BinStr},
         genGotoOutput(Tail, NewGoto, NewOutput, NewMaxState);
      _ ->
         genGotoOutput(Tail, Goto, Output, MaxState)
   end;
genGotoOutput([], Goto, Output, _MaxState) ->
   {Goto, Output}.

addPattern(<<Word/utf8, Tail/binary>>, Goto, State, MaxState) ->
   #{State := Node} = Goto,
   case Node of
      #{Word := NextState} ->
         addPattern(Tail, Goto, NextState, MaxState);
      _ ->
         NewMaxState = MaxState + 1,
         NewNode = Node#{Word => NewMaxState},
         addPattern(Tail, Goto#{NewMaxState => #{}, State => NewNode}, NewMaxState, NewMaxState)
   end;
addPattern(<<>>, Goto, State, MaxState) ->
   {Goto, State, MaxState}.

genFailure(#{0 := Node} = Goto) ->
   States = maps:values(Node),
   genFailure(States, Goto, _Failure = #{}).

%% 构造 failure with bfs搜索
genFailure([State | Tail], Goto, Failure) ->
   #{State := Node} = Goto,
   %% find the starting point: the parent's failure node
   FailState = maps:get(State, Failure, 0),

   %% children
   Kvs = maps:to_list(Node),

   %% find failure node for all children
   NewFailure = genFailureInner(Kvs, FailState, Goto, Failure),

   %% add children states to the queue
   NewQueue = Tail ++ maps:values(Node),

   genFailure(NewQueue, Goto, NewFailure);
genFailure([], _Goto, Failure) ->
   Failure.


%% 为节点构造失败指针
%% @param FailState 是当前节点的失败指针
genFailureInner([], _FailState, _Goto, Failure) ->
   Failure;
genFailureInner([{Word, State} | Tail], FailState, Goto, Failure) ->
   NewFailure = findFailureNode(Word, State, FailState, Goto, Failure),
   genFailureInner(Tail, FailState, Goto, NewFailure).

%% 为某个儿子节点构造失败指针
findFailureNode(Word, State, FailState, Goto, Failure) ->
   #{FailState := Node} = Goto,
   case Node of
      #{Word := TheFailState} ->
         %% 找到最近的失败节点的儿子节点拥有当前儿子节点的值，查找成功
         Failure#{State => TheFailState};
      _ ->
         case FailState =:= 0 of
            true -> %% 找不到，而且已经到了根节点，查找失败
               Failure;
            _ -> %% 找不到但是还没到根节点，继续往上找
               NewFailState = maps:get(FailState, Failure, 0),
               findFailureNode(Word, State, NewFailState, Goto, Failure)
         end
   end.

genHead() ->
   <<"-module(acsTrees).\n\n-compile([deterministic, no_line_info]).\n\n-export([goto/1, failOut/1]).\n\n">>.

genGoto(Goto, StrAcc) ->
   Kvs = maps:to_list(Goto),
   SortKvs = lists:sort(Kvs),
   doGenGoto(SortKvs, StrAcc).

doGenGoto([], StrAcc) ->
   <<StrAcc/binary, "goto(_) -> undefined.\n\n">>;
doGenGoto([{K, V}], StrAcc) ->
   case maps:size(V) > 0 of
      true ->
         <<StrAcc/binary, "goto(", (integer_to_binary(K))/binary, ") -> ", (eFmt:formatBin(<<"~w">>, [V]))/binary, ";\ngoto(_) -> undefined.\n\n">>;
      _ ->
         <<StrAcc/binary, "goto(_) -> undefined.\n\n">>
   end;
doGenGoto([{K, V} | SortKvs], StrAcc) ->
   case maps:size(V) > 0 of
      true ->
         NewStrAcc = <<StrAcc/binary, "goto(", (integer_to_binary(K))/binary, ") -> ", (eFmt:formatBin(<<"~w">>, [V]))/binary, ";\n">>,
         doGenGoto(SortKvs, NewStrAcc);
      _ ->
         doGenGoto(SortKvs, StrAcc)
   end.

genFailOut([], _Failure, _Output, StrAcc) ->
   <<StrAcc/binary, "\nfailOut(_) -> {0, undefined}.\n\n">>;
genFailOut([State], Failure, Output, StrAcc) ->
   FailState = maps:get(State, Failure, 0),
   Pattern = maps:get(State, Output, undefined),
   case FailState /= 0 orelse Pattern /= undefined of
      true ->
         <<StrAcc/binary, "failOut(", (integer_to_binary(State))/binary, ") -> ", (eFmt:formatBin(<<"~w">>, [{FailState, Pattern}]))/binary, ";\nfailOut(_) -> {0, undefined}.\n\n">>;
      _ ->
         <<StrAcc/binary, ";\nfailOut(_) -> {0, undefined}.\n\n">>
   end;
genFailOut([State | SortStates], Failure, Output, StrAcc) ->
   FailState = maps:get(State, Failure, 0),
   Pattern = maps:get(State, Output, undefined),
   case FailState /= 0 orelse Pattern /= undefined of
      true ->
         NewStrAcc = <<StrAcc/binary, "failOut(", (integer_to_binary(State))/binary, ") -> ", (eFmt:formatBin(<<"~w">>, [{FailState, Pattern}]))/binary, ";\n">>,
         genFailOut(SortStates, Failure, Output, NewStrAcc);
      _ ->
         genFailOut(SortStates, Failure, Output, StrAcc)
   end.

genErl(WriteDir, Goto, Failure, Output) ->
   HeadStr = genHead(),
   GotoStr = genGoto(Goto, HeadStr),
   FailureStr = genFailOut(lists:sort(maps:keys(Goto)), Failure, Output, GotoStr),
   FileName = filename:join([WriteDir, "acsTrees.erl"]),
   file:write_file(FileName, FailureStr).



